# Copyright © 2023 Gurobi Optimization, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Define generic function that can add any known trained predictor."""

from .exceptions import NotRegistered
from .modeling.get_convertor import get_convertor
from .registered_predictors import registered_predictors


def add_predictor_constr(gp_model, predictor, input_vars, output_vars=None, **kwargs):
    """Formulate predictor in gp_model.

    The formulation predicts the values of output_vars using input_vars according to
    predictor.

    Parameters
    ----------
    gp_model : :gurobipy:`model`
            The gurobipy model where the predictor should be inserted.
    predictor:
        The predictor to insert.
    input_vars : mvar_array_like
        Decision variables used as input for predictor in gp_model.
    output_vars : mvar_array_like, optional
        Decision variables used as output for predictor in gp_model.

    Returns
    -------
    AbstractPredictorConstr
        Object containing information about what was added to gp_model to insert the
        predictor in it

    Note
    ----
    The parameters `input_vars` and `output_vars` can be either

     * Gurobipy matrix variables :gurobipy:`mvar`
     * Pandas data frames containing columns of variables or constants
     * Lists of variables
     * Dictionaries of variables

    For internal use in the package they are cast into matrix variables.

    They should have dimensions that conforms with the input/output of the predictor.
    We denote by `n_features` the dimension of the input of the predictor and by
    `n_output` the dimension of the output.

    If they are matrix variables, `input_vars` and `output_vars` can be either of
    shape `(n_features)` and `(n_outputs,)` respectively or `(k, n_features)` and
    `(k, n_outputs)` respectively (with `k >= 1`). The latter form is especially
    useful if the predictor is used to associate different groups of variables
    (e.g. a prediction is made for every time period in a planning horizon).

    If they are pandas dataframe, `input_vars` should have the features as columns and
    `output_vars` the outputs of predictors. Note that the input_vars dataframe may
    have *fixed* columns containing constant values and *variable* columns containing
    gurobipy variables. A column should not mix constants and variables.

    If they are lists or dictionaries, `input_vars` should have length `n_features` and
    `output_vars` should have length `n_output`.
    List and dictionaries can have bad performances in particular if particular adding
    many predictor constraints in a for loop is significantly slower.
    """
    convertors = registered_predictors()
    convertor = get_convertor(predictor, convertors)
    if convertor is None:
        raise NotRegistered(type(predictor).__name__)
    return convertor(gp_model, predictor, input_vars, output_vars, **kwargs)



# Copyright © 2023 Gurobi Optimization, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""
Gurobi Machine learning
==================================

A Python package to help use trained regression models in
mathematical optimization models. The package supports a variety of regression models
(linear, logistic, neural networks, decision trees,...) trained by
different machine learning frameworks (scikit-learn, Keras and PyTorch).

See https://gurobi-optimization-gurobi-machine-learning.readthedocs-hosted.com/
for documentation.
"""

# read version from installed package

from ._version import __version__
from .add_predictor import add_predictor_constr
from .register_user_predictor import register_predictor_constr



# Copyright © 2023 Gurobi Optimization, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

""" Utilities for modeling decision trees """

from warnings import warn

import numpy as np
from gurobipy import GRB


def compute_leafs_bounds(tree, epsilon):
    """Compute the bounds that define each leaf of the tree"""
    capacity = tree["capacity"]
    n_features = tree["n_features"]
    children_left = tree["children_left"]
    children_right = tree["children_right"]
    feature = tree["feature"]
    threshold = tree["threshold"]

    node_lb = -np.ones((n_features, capacity)) * GRB.INFINITY
    node_ub = np.ones((n_features, capacity)) * GRB.INFINITY

    stack = [
        0,
    ]

    while len(stack) > 0:
        node = stack.pop()
        left = children_left[node]
        if left < 0:
            continue
        right = children_right[node]
        assert left not in stack
        assert right not in stack
        node_ub[:, right] = node_ub[:, node]
        node_lb[:, right] = node_lb[:, node]
        node_ub[:, left] = node_ub[:, node]
        node_lb[:, left] = node_lb[:, node]

        node_ub[feature[node], left] = threshold[node]
        node_lb[feature[node], right] = threshold[node] + epsilon
        stack.append(right)
        stack.append(left)
    return (node_lb, node_ub)


def leaf_formulation(
    gp_model, _input, output, tree, epsilon, _name_var, verbose, timer
):
    """Formulate decision tree using 'leaf' formulation

    We have one variable per leaf of the tree and a series of indicator to
    define when that leaf is reached.
    """
    nex = _input.shape[0]
    n_features = tree["n_features"]

    # Collect leaf nodes
    leafs = tree["children_left"] < 0
    leafs_vars = gp_model.addMVar(
        (nex, sum(leafs)), vtype=GRB.BINARY, name=_name_var("leafs")
    )

    if verbose:
        timer.timing(f"Added {nex*sum(leafs)} leafs vars")
    (node_lb, node_ub) = compute_leafs_bounds(tree, epsilon)
    input_ub = _input.getAttr(GRB.Attr.UB)
    input_lb = _input.getAttr(GRB.Attr.LB)

    for i, node in enumerate(leafs.nonzero()[0]):
        reachable = (input_ub >= node_lb[:, node]).all(axis=1) & (
            input_lb <= node_ub[:, node]
        ).all(axis=1)
        # Non reachable nodes
        leafs_vars[~reachable, i].setAttr(GRB.Attr.UB, 0.0)
        # Leaf node:
        rhs = output[reachable, :].tolist()
        lhs = leafs_vars[reachable, i].tolist()
        values = tree["value"][node, :]
        n_indicators = sum(reachable)
        for l_var, r_vars in zip(lhs, rhs):
            for r_var, value in zip(r_vars, values):
                gp_model.addGenConstrIndicator(l_var, 1, r_var, GRB.EQUAL, value)

        for feature in range(n_features):
            feat_lb = node_lb[feature, node]
            feat_ub = node_ub[feature, node]

            if feat_lb > -GRB.INFINITY:
                tight = (input_lb[:, feature] < feat_lb) & reachable
                lhs = leafs_vars[tight, i].tolist()
                rhs = _input[tight, feature].tolist()
                n_indicators += sum(tight)
                for l_var, r_var in zip(lhs, rhs):
                    gp_model.addGenConstrIndicator(
                        l_var, 1, r_var, GRB.GREATER_EQUAL, feat_lb
                    )

            if feat_ub < GRB.INFINITY:
                tight = (input_ub[:, feature] > feat_ub) & reachable
                lhs = leafs_vars[tight, i].tolist()
                rhs = _input[tight, feature].tolist()
                n_indicators += sum(tight)
                for l_var, r_var in zip(lhs, rhs):
                    gp_model.addGenConstrIndicator(
                        l_var, 1, r_var, GRB.LESS_EQUAL, feat_ub
                    )
        if verbose:
            timer.timing(f"Added leaf {node} using {n_indicators} indicators")

    # We should attain 1 leaf
    gp_model.addConstr(leafs_vars.sum(axis=1) == 1)

    if verbose:
        timer.timing(f"Added {nex} linear constraints")

    output.setAttr(GRB.Attr.LB, np.min(tree["value"]))
    output.setAttr(GRB.Attr.UB, np.max(tree["value"]))


def paths_formulation(gp_model, _input, output, tree, epsilon, _name_var):
    """
       Path formulation for decision tree

    We have one variable for each node of the tree and do a formulation
    that reconsistutes paths through the tree. This is inferior to the
    leaf formulation and is deprecated.
    """

    warn(
        "Path formulation of decision trees is not tested anymore.", DeprecationWarning
    )
    outdim = output.shape[1]
    nex = _input.shape[0]
    nodes = gp_model.addMVar(
        (nex, tree.capacity), vtype=GRB.BINARY, name=_name_var("node")
    )

    children_left = tree["children_left"]
    children_right = tree["children_right"]
    threshold = tree["threshold"]
    feature = tree["feature"]
    value = tree["value"]
    # Collect leafs and non-leafs nodes
    not_leafs = children_left >= 0
    leafs = children_left < 0

    # Connectivity constraint
    gp_model.addConstr(
        nodes[:, not_leafs]
        == nodes[:, children_right[not_leafs]] + nodes[:, children_left[not_leafs]]
    )

    # The value of the root is always 1
    nodes[:, 0].LB = 1.0

    # Node splitting
    for node in not_leafs.nonzero()[0]:
        left = children_left[node]
        right = children_right[node]
        node_threshold = threshold[node]
        # Intermediate node
        node_feature = feature[node]
        feat_var = _input[:, node_feature]

        fixed_input = (feat_var.UB == feat_var.LB).all()

        if fixed_input:
            # Special case where we have an MVarPlusConst object
            # If that feature is a constant we can directly fix it.
            node_value = _input[:, node_feature].LB
            fixed_left = node_value <= node_threshold
            nodes[fixed_left, right].UB = 0.0
            nodes[~fixed_left, left].UB = 0.0
        else:
            lhs = _input[:, feature].tolist()
            rhs = nodes[:, left].tolist()
            gp_model.addConstrs(
                ((rhs[k] == 1) >> (lhs[k] <= node_threshold)) for k in range(nex)
            )
            rhs = nodes[:, right].tolist()
            gp_model.addConstrs(
                ((rhs[k] == 1) >> (lhs[k] >= node_threshold + epsilon))
                for k in range(nex)
            )

    for node in leafs.nonzero()[0]:
        # Leaf node:
        lhs = output.tolist()
        rhs = nodes[:, node].tolist()
        node_value = value[node, :]
        gp_model.addConstrs(
            (rhs[k] == 1) >> (lhs[k][i] == node_value[i])
            for k in range(nex)
            for i in range(outdim)
        )

    output.LB = np.min(tree.value)
    output.UB = np.max(tree.value)



# Copyright © 2023 Gurobi Optimization, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Implements some utility tools for all scikit-learn objects."""

import numpy as np
from sklearn.utils.validation import check_is_fitted

from ..exceptions import NoSolution
from ..modeling import AbstractPredictorConstr


class SKgetter(AbstractPredictorConstr):
    """Utility class for sklearn regression models convertors.

    Implement some common functionalities: check predictor is fitted, output dimension, get error

    Attributes
    ----------
    predictor
        Scikit-Learn predictor embedded into Gurobi model.
    """

    def __init__(self, predictor, input_vars, output_type="regular", **kwargs):
        check_is_fitted(predictor)
        self.predictor = predictor
        predictor._check_feature_names(input_vars, reset=False)
        self.output_type = output_type
        if hasattr(predictor, "n_features_in_"):
            self._input_shape = predictor.n_features_in_
        if hasattr(predictor, "n_outputs_"):
            self._output_shape = predictor.n_outputs_

    def get_error(self, eps=None):
        """Return error in Gurobi's solution with respect to prediction from input.

        Returns
        -------
        error : ndarray of same shape as :py:attr:`gurobi_ml.modeling.base_predictor_constr.AbstractPredictorConstr.output`
            Assuming that we have a solution for the input and output variables
            `x, y`. Returns the absolute value of the differences between `predictor.predict(x)` and
            `y`. Where predictor is the regression this object is modeling.

        Raises
        ------
        NoSolution
            If the Gurobi model has no solution (either was not optimized or is infeasible).
        """
        if self._has_solution:
            X = self.input_values
            if self.output_type == "probability_1":
                predicted = self.predictor.predict_proba(X)[:, 1]
            elif self.output_type == "probability":
                predicted = self.predictor.predict_proba(X)
            else:
                predicted = self.predictor.predict(X)
            output_values = self.output_values
            if len(predicted.shape) == 1 and len(output_values.shape) == 2:
                predicted = predicted.reshape(-1, 1)
            r_val = np.abs(predicted - output_values)
            if eps is not None and np.max(r_val) > eps:
                print(f"{predicted} != {output_values}")
            return r_val

        raise NoSolution()


class SKtransformer(AbstractPredictorConstr):
    """Utility class for sklearn preprocessing models convertors.

    Implement some common functionalities.

    Attributes
    ----------
    transformer
        Scikit-Learn transformer embedded into Gurobi model.
    """

    def __init__(self, gp_model, transformer, input_vars, **kwargs):
        self.transformer = transformer
        if hasattr(transformer, "n_features_in_"):
            self._input_shape = transformer.n_features_in_
        if hasattr(transformer, "n_output_features_"):
            self._output_shape = transformer.n_output_features_
        check_is_fitted(transformer)
        super().__init__(gp_model, input_vars, **kwargs)

    def get_error(self, eps=None):
        """Return error in Gurobi's solution with respect to preprocessing from input.

        Returns
        -------
        error : ndarray of same shape as :py:attr:`gurobi_ml.modeling.base_predictor_constr.AbstractPredictorConstr.output`
            Assuming that we have a solution for the input and output variables
            `x, y`. Returns the absolute value of the differences between `transformer.transform(x)` and
            `y`. Where transformer is the prepocessing this object is modeling.

        Raises
        ------
        NoSolution
            If the Gurobi model has no solution (either was not optimized or is infeasible).
        """
        if self._has_solution:
            transformer = self.transformer
            input_values = self.input_values

            transformed = transformer.transform(input_values)
            if len(transformed.shape) == 1:
                transformed = transformed.reshape(-1, 1)

            r_val = np.abs(transformed - self.output_values)
            if eps is not None and np.max(r_val) > eps:
                print(f"{transformed} != {self.output_values}")
            return r_val

        raise NoSolution()



# Copyright © 2023 Gurobi Optimization, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Module for formulating simple Scikit-Learn PLS regression models in a gurobipy model."""

import numpy as np

from ..modeling import AbstractPredictorConstr
from .skgetter import SKgetter


def add_pls_regression_constr(
    gp_model, pls_regression, input_vars, output_vars=None, **kwargs
):
    """Formulate pls_regression in gp_model.

    The formulation predicts the values of output_vars using input_vars
    according to pls_regression.

    Parameters
    ----------
    gp_model : :gurobipy:`model`
        The gurobipy model where the predictor should be inserted.
    pls_regression : :external+sklearn:py:class:`sklearn.cross_decomposition.PLSRegression`
     The linear regression to insert. It can be of any of the following types:
         * :external+sklearn:py:class:`sklearn.cross_decomposition.PLSRegression`
         * :external+sklearn:py:class:`sklearn.cross_decomposition.PLSCanonical`
     input_vars: :gurobipy:`mvar` or :gurobipy:`var` array like
         Decision variables used as input for random forest in model.
     output_vars: :gurobipy:`mvar` or :gurobipy:`var` array like, optional
         Decision variables used as output for random forest in model.

    Returns
    -------
    PLSRegressionConstr
        Object containing information about what was added to gp_model to
        formulate pls_regression.

    Note
    ----
    |VariablesDimensionsWarn|
    """
    return PLSRegressionConstr(
        gp_model, pls_regression, input_vars, output_vars, **kwargs
    )


class PLSRegressionConstr(SKgetter, AbstractPredictorConstr):
    """Class to model trained
    :external+sklearn:py:class:`sklearn.cross_decomposition.PLSRegression` with
    gurobipy
    |ClassShort|.
    """

    def __init__(
        self,
        gp_model,
        predictor,
        input_vars,
        output_vars=None,
        output_type="",
        **kwargs,
    ):
        self._output_shape = 1
        SKgetter.__init__(self, predictor, input_vars, output_type, **kwargs)
        AbstractPredictorConstr.__init__(
            self,
            gp_model,
            input_vars,
            output_vars,
            **kwargs,
        )

    def add_regression_constr(self):
        """Add the prediction constraints to Gurobi."""
        x_mean = self.predictor._x_mean
        x_std = self.predictor._x_std
        coefs = self.predictor._coef_.T
        intercept = self.predictor.intercept_
        self.gp_model.addConstr(
            self.output
            == self.input @ (coefs / x_std[:, np.newaxis])
            - x_mean / x_std @ coefs
            + intercept,
            name="plsreg",
        )

    def _mip_model(self, **kwargs):
        """Add the prediction constraints to Gurobi."""
        self.add_regression_constr()



# Copyright © 2023 Gurobi Optimization, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Module for formulating a :external+sklearn:py:class:`sklearn.pipeline.Pipeline`
in a :gurobipy:`model`.
"""


from ..exceptions import NoModel
from ..modeling.base_predictor_constr import AbstractPredictorConstr
from ..modeling.get_convertor import get_convertor
from ..register_user_predictor import user_predictors
from .column_transformer import add_column_transformer_constr
from .predictors_list import sklearn_predictors
from .preprocessing import sklearn_transformers
from .skgetter import SKgetter


def add_pipeline_constr(gp_model, pipeline, input_vars, output_vars=None, **kwargs):
    """Formulate pipeline into gp_model.

    The formulation predicts the values of output_vars using input_vars according to
    pipeline.

    Parameters
    ----------
    gp_model : :gurobipy:`model`
        The gurobipy model where the predictor should be inserted.
    pipeline : :external+sklearn:py:class:`sklearn.pipeline.Pipeline`
        The pipeline to insert as predictor.
    input_vars : :gurobipy:`mvar` or :gurobipy:`var` array like
        Decision variables used as input for regression in model.
    output_vars : :gurobipy:`mvar` or :gurobipy:`var` array like, optional
        Decision variables used as output for regression in model.

    Returns
    -------
    PipelineConstr
        Object containing information about what was added to gp_model to embed the
        predictor into it

    Raises
    ------
    NoModel
        If the translation to Gurobi of one of the elements in the pipeline
        is not implemented or recognized.

    Note
    ----
    |VariablesDimensionsWarn|
    """
    return PipelineConstr(gp_model, pipeline, input_vars, output_vars, **kwargs)


class PipelineConstr(SKgetter, AbstractPredictorConstr):
    """Class to model trained :external+sklearn:py:class:`sklearn.pipeline.Pipeline`
    with gurobipy
    |ClassShort|.
    """

    def __init__(self, gp_model, pipeline, input_vars, output_vars=None, **kwargs):
        self._steps = []
        self._default_name = "pipe"
        SKgetter.__init__(self, pipeline, input_vars, **kwargs)
        AbstractPredictorConstr.__init__(
            self, gp_model, input_vars, output_vars, validate_input=False, **kwargs
        )

    def _build_submodel(self, gp_model, *args, **kwargs):
        """Predict output from input using predictor or transformer.

        Pipelines are different from other objects because they can't validate
        their input and output. They are just containers of other objects that will
        do it.
        """
        self._mip_model(**kwargs)
        assert self.output is not None
        assert self.input is not None
        # We can call validate only after the model is created
        self._validate()
        return self

    def _mip_model(self, **kwargs):
        pipeline = self.predictor
        gp_model = self._gp_model
        input_vars = self._input
        output_vars = self._output
        steps = self._steps
        transformers = sklearn_transformers()
        transformers |= user_predictors()
        transformers["ColumnTransformer"] = add_column_transformer_constr
        kwargs["validate_input"] = True

        for transformer in pipeline[:-1]:
            convertor = get_convertor(transformer, transformers)
            if convertor is None:
                raise NoModel(
                    self.predictor,
                    f"I don't know how to deal with that object: {transformer}",
                )
            steps.append(convertor(gp_model, transformer, input_vars, **kwargs))
            input_vars = steps[-1].output

        predictor = pipeline[-1]
        predictors = sklearn_predictors() | user_predictors()
        convertor = get_convertor(predictor, predictors)
        if convertor is None:
            raise NoModel(
                self.predictor,
                f"I don't know how to deal with that object: {predictor}",
            )
        steps.append(convertor(gp_model, predictor, input_vars, output_vars, **kwargs))
        if self._output is None:
            self._output = steps[-1].output

    def print_stats(self, file=None):
        """Print statistics on model additions stored by this class.

        This function prints detailed statistics on the variables
        and constraints that where added to the model.

        The pipeline version includes a summary of the steps that it contains.

        Arguments
        ---------

        file: None, optional
            Text stream to which output should be redirected. By default sys.stdout.
        """
        super().print_stats(file=file)
        print(file=file)
        print(f"Pipeline has {len(self._steps)} steps:", file=file)
        print(file=file)

        self._print_container_steps("Step", self._steps, file=file)

    @property
    def _has_solution(self):
        return self[-1]._has_solution

    @property
    def output(self):
        """Returns output variables of pipeline, i.e. output of its last step."""
        return self[-1].output

    @property
    def output_values(self):
        """Returns output values of pipeline in solution, i.e. output of its last step."""
        return self[-1].output_values

    @property
    def input(self):
        """Returns input variables of pipeline, i.e. input of its first step."""
        return self[0].input

    @property
    def input_values(self):
        """Returns input values of pipeline in solution, i.e. input of its first step."""
        return self[0].input_values

    def __getitem__(self, key):
        """Get an item from the pipeline steps."""
        return self._steps[key]

    def __iter__(self):
        """Iterate through pipeline steps."""
        return self._steps.__iter__()

    def __len__(self):
        """Get number of pipeline steps."""
        return self._steps.__len__()



# Copyright © 2023 Gurobi Optimization, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Bases classes for modeling neural network layers."""

import io

import gurobipy as gp

from gurobi_ml.modeling.neuralnet.activations import Identity

from .._var_utils import _default_name
from ..base_predictor_constr import AbstractPredictorConstr


class AbstractNNLayer(AbstractPredictorConstr):
    """Abstract class for NN layers."""

    def __init__(
        self,
        gp_model,
        output_vars,
        input_vars,
        activation_function,
        **kwargs,
    ):
        self.activation = activation_function
        AbstractPredictorConstr.__init__(
            self, gp_model, input_vars, output_vars, **kwargs
        )

    def get_error(self, eps=None):
        # We can't compute externally the error of a layer
        assert False

    def print_stats(self, abbrev=False, file=None):
        """Print statistics about submodel created.

        Parameters
        ----------

        file : None, optional
          Text stream to which output should be redirected. By default sys.stdout.
        """
        return AbstractPredictorConstr.print_stats(self, True, file)


class ActivationLayer(AbstractNNLayer):
    """Class to build one activation layer of a neural network."""

    def __init__(
        self,
        gp_model,
        output_vars,
        input_vars,
        activation_function,
        **kwargs,
    ):
        self.zvar = None
        self._default_name = "activation"
        super().__init__(
            gp_model,
            output_vars,
            input_vars,
            activation_function,
            **kwargs,
        )

    def _create_output_vars(self, input_vars):
        rval = self._gp_model.addMVar(input_vars.shape, lb=-gp.GRB.INFINITY, name="act")
        self._gp_model.update()
        return rval

    def _mip_model(self, **kwargs):
        """Add the layer to model."""
        model = self.gp_model
        model.update()
        if "activation" in kwargs:
            activation = kwargs["activation"]
        else:
            activation = self.activation

        # Do the mip model for the activation in the layer
        activation.mip_model(self)
        self._gp_model.update()


class DenseLayer(AbstractNNLayer):
    """Class to build one layer of a neural network."""

    def __init__(
        self,
        gp_model,
        output_vars,
        input_vars,
        layer_coefs,
        layer_intercept,
        activation_function,
        **kwargs,
    ):
        self.coefs = layer_coefs
        self.intercept = layer_intercept
        self.zvar = None
        self._default_name = "dense"
        super().__init__(
            gp_model,
            output_vars,
            input_vars,
            activation_function,
            **kwargs,
        )

    def _create_output_vars(self, input_vars):
        rval = self._gp_model.addMVar(
            (input_vars.shape[0], self.coefs.shape[1]), lb=-gp.GRB.INFINITY, name="act"
        )
        self._gp_model.update()
        return rval

    def _mip_model(self, **kwargs):
        """Add the layer to model."""
        model = self.gp_model
        model.update()
        if "activation" in kwargs:
            activation = kwargs["activation"]
        else:
            activation = self.activation

        # Do the mip model for the activation in the layer
        activation.mip_model(self)
        self._gp_model.update()

    def print_stats(self, abbrev=False, file=None):
        """Print statistics about submodel created.

        Parameters
        ----------

        file : None, optional
          Text stream to which output should be redirected. By default sys.stdout.
        """
        if not isinstance(self.activation, Identity):
            output = io.StringIO()
            AbstractPredictorConstr.print_stats(self, abbrev=True, file=output)
            activation_name = f"({_default_name(self.activation)})"

            out_string = output.getvalue()
            print(f"{out_string[:-1]} {activation_name}", file=file)
            return
        AbstractPredictorConstr.print_stats(self, abbrev=True, file=file)



# Copyright © 2023 Gurobi Optimization, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Utility function to find function that add a predictor in dictionnary."""


def get_convertor(predictor, convertors):
    """Return the convertor for a given predictor."""
    convertor = None
    try:
        convertor = convertors[type(predictor)]
    except KeyError:
        pass
    if convertor is None:
        for parent in type(predictor).mro():
            try:
                convertor = convertors[parent]
                break
            except KeyError:
                pass
    if convertor is None:
        name = type(predictor).__name__
        try:
            convertor = convertors[name]
        except KeyError:
            pass
    return convertor



# Copyright © 2023 Gurobi Optimization, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Module for formulating a
:external+sklearn:py:class:`sklearn.tree.DecisionTreeRegressor`
in a :gurobipy:`model`.
"""


from ..modeling import AbstractPredictorConstr
from ..modeling.decision_tree.decision_tree_model import (
    leaf_formulation,
    paths_formulation,
)
from .skgetter import SKgetter


def add_decision_tree_regressor_constr(
    gp_model,
    decision_tree_regressor,
    input_vars,
    output_vars=None,
    epsilon=0.0,
    **kwargs,
):
    """Formulate decision_tree_regressor into gp_model.

    The formulation predicts the values of output_vars using input_vars
    according to decision_tree_regressor. See our :ref:`User's Guide <Decision
    Tree Regression>` for details on the mip formulation used.

    Parameters
    ----------
    gp_model : :gurobipy:`model`
        The gurobipy model where the predictor should be inserted.
    decision_tree_regressor : :external+sklearn:py:class:`sklearn.tree.DecisionTreeRegressor`
        The decision tree regressor to insert as predictor.
    input_vars : :gurobipy:`mvar` or :gurobipy:`var` array like
        Decision variables used as input for decision tree in model.
    output_vars : :gurobipy:`mvar` or :gurobipy:`var` array like, optional
        Decision variables used as output for decision tree in model.
    epsilon : float, optional
        Small value used to impose strict inequalities for splitting nodes in
        MIP formulations.
    Returns
    -------
    DecisionTreeRegressorConstr
        Object containing information about what was added to gp_model to
        formulate decision_tree_regressor

    Note
    ----

    |VariablesDimensionsWarn|

    Warning
    -------

    Although decision trees with multiple outputs are tested they were never
    used in a non-trivial optimization model. It should be used with care at
    this point.
    """
    return DecisionTreeRegressorConstr(
        gp_model, decision_tree_regressor, input_vars, output_vars, epsilon, **kwargs
    )


class DecisionTreeRegressorConstr(SKgetter, AbstractPredictorConstr):
    """Class to model trained
    :external+sklearn:py:class:`sklearn.tree.DecisionTreeRegressor` with
    gurobipy.

    |ClassShort|
    """

    def __init__(
        self,
        gp_model,
        predictor,
        input_vars,
        output_vars=None,
        epsilon=0.0,
        formulation="leafs",
        **kwargs,
    ):
        self.epsilon = epsilon
        self._default_name = "tree_reg"

        formulations = ("leafs", "paths")
        if formulation not in formulations:
            raise ValueError(
                "Wrong value for formulation should be one of {}.".format(formulations)
            )
        self._formulation = formulation
        SKgetter.__init__(self, predictor, input_vars)
        AbstractPredictorConstr.__init__(
            self, gp_model, input_vars, output_vars, **kwargs
        )

    def _mip_model(self, **kwargs):
        tree = self.predictor.tree_
        timer = AbstractPredictorConstr._ModelingTimer()

        tree_dict = {
            "children_left": tree.children_left,
            "children_right": tree.children_right,
            "feature": tree.feature,
            "threshold": tree.threshold,
            "value": tree.value[:, :, 0],
            "capacity": tree.capacity,
            "n_features": tree.n_features,
        }
        if self._formulation == "leafs":
            return leaf_formulation(
                self.gp_model,
                self.input,
                self.output,
                tree_dict,
                self.epsilon,
                self._name_var,
                self.verbose,
                timer,
            )
        else:
            return paths_formulation(
                self.gp_model,
                self.input,
                self.output,
                tree_dict,
                self.epsilon,
                self._name_var,
            )



# Copyright © 2023 Gurobi Optimization, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Utility functions for dealing with input and output variables of predictor constr objects."""

import gurobipy as gp
import numpy as np

from ..exceptions import ParameterError

try:
    import pandas as pd

    HAS_PANDAS = True
except ImportError:
    HAS_PANDAS = False


def _default_name(predictor):
    """Make a default name for predictor constraint.

    Parameters
    ----------
    predictor:
        Class of the predictor
    """
    return type(predictor).__name__.lower()


def _get_sol_values(values, columns=None, index=None):
    """Get solution values.

    This is complicated because of the column_transformer.
    In most cases we can just do values.X but if we have a column transformer with
    some constants that can't be translated to Gurobi variables we need to fill in missing values
    """
    if HAS_PANDAS:
        if isinstance(values, pd.DataFrame):
            rval = pd.DataFrame(
                data=_get_sol_values(values.to_numpy()),
                index=values.index,
                columns=values.columns,
            )
            for col in rval.columns:
                try:
                    rval[col] = rval[col].astype(np.float64)
                except ValueError:
                    pass
            return rval.convert_dtypes()
    if isinstance(values, np.ndarray):
        return np.array(
            [v.X if isinstance(v, gp.Var) else v for v in values.ravel()]
        ).reshape(values.shape)
    X = values.X
    if columns is not None and HAS_PANDAS:
        if isinstance(index, (pd.Index, pd.MultiIndex)):
            X = pd.DataFrame(data=X, columns=columns, index=index)
        else:
            X = pd.Series(data=X, columns=columns, name=index)
            raise NotImplementedError("Input variables as pd.Series is not implemented")
    return X


def _dataframe_to_mvar(model, df):
    """Convert a DataFrame to an mvar.

    This just calls the array_to_mvar function
    """
    if isinstance(df, pd.DataFrame):
        data = df.to_numpy()
        columns = df.columns
        index = df.index
    elif isinstance(df, pd.Series):
        raise NotImplementedError("Input variables as pd.Series is not implemented")
    return _array_to_mvar(model, data, columns, index)


def _array_to_mvar(model, data, columns=None, index=None):
    """Function to create an MVar from an ndarray.

    The array data may contain columns of Gurobi variables and constant values.
    In this function we create an array of same shape as data where constant
    values are replace by fixed variables. This is then converted to an MVar.

    If columns and index are passed, used them to give appropriate names to
    created variables.

    Parameters
    ----------
    model : gp.Model
        The gurobipy model
    data : ndarray
        A numpy array whose columns are either columns of gp.Var or data that
        can be converted to a float
    columns : optional
        Name of data columns
    index:
        Name of data rows
    """

    # If data only contains gp.Var's we can directly convert it to an MVar
    if all(map(lambda i: isinstance(i, gp.Var), data.ravel())):
        rval = gp.MVar.fromlist(data.tolist())
        return rval

    # data doesn't contain only contain gp.Var,s.
    # Iterate through the columns, keep columns of variables as is.
    # For other columns check if they can be converted to numbers and collect
    # their indices
    rval = np.zeros(data.shape, dtype=object)
    const_indices = []
    for i, a in enumerate(data.T):
        if all(map(lambda i: isinstance(i, gp.Var), a)):
            rval[:, i] = a
            continue
        try:
            rval[:, i] = a.astype(np.float64)
        except TypeError:
            raise ValueError(
                f"Column {i} of input variables can't be converted to gurobipy variables or floats"
            )
        const_indices.append(i)

    # If columns and index are not passed try to infer something to name the variables
    if columns is None or index is None:
        if data.ndim == 2:
            main_shape = data.shape[1]
            minor_shape = data.shape[0]
        else:
            main_shape = data.shape[0]
            minor_shape = 1
        if columns is None:
            columns = [f"feat{i}" for i in range(main_shape)]
        if index is None:
            index = list(range(minor_shape))

    # Create missing variables, fix them and put them into the array
    mvar = model.addMVar((data.shape[0], len(const_indices)))
    for i, j in enumerate(const_indices):
        mvar[:, i].LB = rval[:, j]
        mvar[:, i].UB = rval[:, j]
        mvar[:, i].VarName = [f"{columns[j]}[{k}]" for k in index]
        rval[:, j] = mvar[:, i].tolist()
    model.update()
    rval = gp.MVar.fromlist(rval)
    return rval


def validate_output_vars(gp_vars):
    """Put variables into appropriate form (matrix of variable) for the output of a predictor constraint.

    Parameters
    ----------
    gpvars:
        Decision variables used.

    Returns
    -------
    mvar_array_like
        Decision variables with correctly adjusted shape.
    """
    if HAS_PANDAS:
        if isinstance(gp_vars, (pd.DataFrame, pd.Series)):
            return validate_output_vars(gp_vars.to_numpy())
    if isinstance(gp_vars, np.ndarray):
        if any(map(lambda i: not isinstance(i, gp.Var), gp_vars.ravel())):
            raise TypeError("Dataframe can't be converted to an MVar")
        rval = gp.MVar.fromlist(gp_vars.tolist())
        return rval
    if isinstance(gp_vars, gp.MVar):
        if gp_vars.ndim in (1, 2):
            return gp_vars
        raise ParameterError("Variables should be an MVar of dimension 1 or 2")
    if isinstance(gp_vars, dict):
        gp_vars = gp_vars.values()
    if isinstance(gp_vars, list):
        return gp.MVar.fromlist(gp_vars)
    if isinstance(gp_vars, gp.Var):
        return gp.MVar.fromlist([gp_vars]).reshape(1, 1)
    raise ParameterError("Could not validate variables")


def validate_input_vars(model, gp_vars):
    """Put variables into appropriate form (matrix of variable) for the input of a predictor constraint.

    Parameters
    ----------
    gpvars:
        Decision variables used.

    Returns
    -------
    mvar_array_like
        Decision variables with correctly adjusted shape.
    """
    if HAS_PANDAS:
        if isinstance(gp_vars, (pd.DataFrame, pd.Series)):
            columns = gp_vars.columns
            index = gp_vars.index
            gp_vars = _dataframe_to_mvar(model, gp_vars)
            return (gp_vars, columns, index)
    if isinstance(gp_vars, np.ndarray):
        return (_array_to_mvar(model, gp_vars), None, None)
    if isinstance(gp_vars, gp.MVar):
        if gp_vars.ndim == 1:
            return (gp_vars.reshape(1, -1), None, None)
        if gp_vars.ndim == 2:
            return (gp_vars, None, None)
        raise ParameterError("Variables should be an MVar of dimension 1 or 2")
    if isinstance(gp_vars, dict):
        gp_vars = gp_vars.values()
    if isinstance(gp_vars, list):
        return (gp.MVar.fromlist(gp_vars).reshape(1, -1), None, None)
    if isinstance(gp_vars, gp.Var):
        return (gp.MVar.fromlist([gp_vars]).reshape(1, 1), None, None)
    raise ParameterError("Could not validate variables")



# Copyright © 2023 Gurobi Optimization, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Module for formulating a
XGBoost gradient boosting regressor
into a :gurobipy:`model`.
"""


import json

import numpy as np
import xgboost as xgb
from gurobipy import GRB

from ..exceptions import NoModel, NoSolution
from ..modeling import AbstractPredictorConstr
from ..modeling.decision_tree import leaf_formulation


def add_xgbrfregressor_constr(
    gp_model, xgboost_regressor, input_vars, output_vars=None, epsilon=0.0, **kwargs
):
    """Formulate xgboost_regressor into gp_model.

    The formulation predicts the values of output_vars using input_vars
    according to xgboost_regressor. See our :ref:`User's Guide
    <Gradient Boosting Regression>` for details on the mip formulation used.

    This version is for using directly with the Scikit-Learn wrapper of XGBoost.
    Note that only "gbtree" regressors are supported at this point.

    Parameters
    ----------
    gp_model : :gurobipy:`model`
        The gurobipy model where the predictor should be inserted.
    xgboost_regressor : :external+xgb:py:class:`xgboost.XGBRFRegressor`
        The gradient boosting regressor to insert as predictor.
    input_vars : :gurobipy:`mvar` or :gurobipy:`var` array like
        Decision variables used as input for gradient boosting regressor in model.
    output_vars : :gurobipy:`mvar` or :gurobipy:`var` array like, optional
        Decision variables used as output for gradient boosting regressor in model.

    Returns
    -------
    XGBoostRegressorConstr
        Object containing information about what was added to gp_model to formulate
        gradient_boosting_regressor.

    Note
    ----
    |VariablesDimensionsWarn|

    Also see
    :py:func:`gurobi_ml.sklearn.decision_tree_regressor.add_decision_tree_regressor`
    for specific parameters to model decision tree estimators.

    Raises
    ------
    NoModel
        If the booster is not of type "gbtree".
    """
    return XGBoostRegressorConstr(
        gp_model,
        xgboost_regressor.get_booster(),
        input_vars,
        output_vars,
        epsilon=epsilon,
        **kwargs,
    )


def add_xgboost_regressor_constr(
    gp_model, xgboost_regressor, input_vars, output_vars=None, epsilon=0.0, **kwargs
):
    """Formulate xgboost_regressor into gp_model.

    The formulation predicts the values of output_vars using input_vars
    according to xgboost_regressor. See our :ref:`User's Guide
    <Gradient Boosting Regression>` for details on the mip formulation used.

    Note that only "gbtree" regressors are supported at this point.

    Parameters
    ----------
    gp_model : :gurobipy:`model`
        The gurobipy model where the predictor should be inserted.
    xgboost_regressor : :external+xgb:py:class:`xgboost.Booster`
        The gradient boosting regressor to insert as predictor.
    input_vars : :gurobipy:`mvar` or :gurobipy:`var` array like
        Decision variables used as input for gradient boosting regressor in model.
    output_vars : :gurobipy:`mvar` or :gurobipy:`var` array like, optional
        Decision variables used as output for gradient boosting regressor in model.

    Returns
    -------
    XGBoostRegressorConstr
        Object containing information about what was added to gp_model to formulate
        gradient_boosting_regressor.

    Note
    ----
    |VariablesDimensionsWarn|

    Also see
    :py:func:`gurobi_ml.sklearn.decision_tree_regressor.add_decision_tree_regressor`
    for specific parameters to model decision tree estimators.

    Raises
    ------
    NoModel
        If the booster is not of type "gbtree".
    """
    return XGBoostRegressorConstr(
        gp_model, xgboost_regressor, input_vars, output_vars, epsilon=epsilon, **kwargs
    )


class TreeEstimator(AbstractPredictorConstr):
    def __init__(
        self, gp_model, tree, input_vars, output_vars, epsilon, timer, verbose, **kwargs
    ):
        self._default_name = "tree"
        self._tree = tree
        self._epsilon = epsilon
        self._timer = timer
        self._verbose = verbose
        AbstractPredictorConstr.__init__(
            self, gp_model, input_vars, output_vars, **kwargs
        )

    def _mip_model(self, **kwargs):
        leaf_formulation(
            self._gp_model,
            self.input,
            self.output,
            self._tree,
            self._epsilon,
            self._name_var,
            self._verbose,
            self._timer,
        )

    def get_error(self, eps):
        assert False


class XGBoostRegressorConstr(AbstractPredictorConstr):
    """Class to model trained :external+xgb:py:class:`xgboost.Booster`
    with gurobipy.

    |ClassShort|
    """

    def __init__(
        self, gp_model, xgb_regressor, input_vars, output_vars, epsilon=0.0, **kwargs
    ):
        self._output_shape = 1
        self.estimators_ = []
        self.xgb_regressor = xgb_regressor
        self._default_name = "xgb_reg"
        self.epsilon = epsilon
        AbstractPredictorConstr.__init__(
            self, gp_model, input_vars, output_vars, **kwargs
        )

    def _mip_model(self, **kwargs):
        """Predict output variables y from input variables X using the
        decision tree.

        Both X and y should be array or list of variables of conforming dimensions.
        """
        model = self._gp_model
        xgb_regressor = self.xgb_regressor

        _input = self._input
        output = self._output
        nex = _input.shape[0]
        timer = AbstractPredictorConstr._ModelingTimer()
        outdim = output.shape[1]
        assert (
            outdim == 1
        ), "Output dimension of gradient boosting regressor should be 1"

        xgb_raw = json.loads(xgb_regressor.save_raw(raw_format="json"))
        booster_type = xgb_raw["learner"]["gradient_booster"]["name"]
        if booster_type != "gbtree":
            raise NoModel(xgb_regressor, f"model not implemented for {booster_type}")
        trees = xgb_raw["learner"]["gradient_booster"]["model"]["trees"]
        n_estimators = len(trees)

        estimators = []
        if self._no_debug:
            kwargs["no_record"] = True

        tree_vars = model.addMVar(
            (nex, n_estimators, 1),
            lb=-GRB.INFINITY,
            name=self._name_var("esimator"),
        )

        for i, tree in enumerate(trees):
            if self.verbose:
                self._timer.timing(f"Estimator {i}")
            tree["threshold"] = (
                np.array(tree["split_conditions"], dtype=np.float32) - self.epsilon
            )
            tree["children_left"] = np.array(tree["left_children"])
            tree["children_right"] = np.array(tree["right_children"])
            tree["feature"] = np.array(tree["split_indices"])
            tree["value"] = tree["threshold"].reshape(-1, 1)
            tree["capacity"] = len(tree["split_conditions"])
            tree["n_features"] = int(tree["tree_param"]["num_feature"])

            def _name_tree_var(name):
                rval = self._name_var(name)
                if rval is None:
                    return None
                return rval + f"_{i}"

            estimators.append(
                TreeEstimator(
                    self.gp_model,
                    tree,
                    self.input,
                    tree_vars[:, i, :],
                    self.epsilon,
                    timer,
                    self.verbose,
                    **kwargs,
                )
            )

        self.estimators_ = estimators

        constant = float(xgb_raw["learner"]["learner_model_param"]["base_score"])
        learning_rate = 1.0
        model.addConstr(output == learning_rate * tree_vars.sum(axis=1) + constant)

    def print_stats(self, abbrev=False, file=None):
        """Print statistics on model additions stored by this class.

        This function prints detailed statistics on the variables
        and constraints that where added to the model.

        Includes a summary of the estimators that it contains.

        Arguments
        ---------

        file: None, optional
            Text stream to which output should be redirected. By default sys.stdout.
        """
        super().print_stats(abbrev=abbrev, file=file)
        if abbrev or self._no_debug:
            return
        print(file=file)

        # self._print_container_steps("Estimator", self.estimators_, file=file)

    def get_error(self, eps=None):
        if self._has_solution:
            xgb_in = xgb.DMatrix(self.input_values)
            xgb_out = self.xgb_regressor.predict(xgb_in)
            r_val = np.abs(xgb_out.reshape(-1, 1) - self.output.X)
            if eps is not None and np.max(r_val) > eps:
                print(f"{self.output.X} != {xgb_out.reshape(-1, 1)}")
            return r_val
        raise NoSolution()
